@@ -8,14 +8,15 @@
  * Author: Peter Jones <pjones@redhat.com>
  */
 
-#include <err.h>
 #include <dirent.h>
 #include <efiboot.h>
 #include <efivar.h>
+#include <err.h>
 #include <errno.h>
 #include <inttypes.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -28,71 +29,25 @@
 #include "ucs2.h"
 #include "fwup-efi.h"
 
-static __thread int __fwup_error;
+static char *arch_names_32[] = {
+#if defined(__x86_64__) || defined(__i386__) || defined(__i686__)
+	"ia32",
+#endif
+	""
+	};
 
-int *
-__fwup_error_location(void)
-{
-	return &__fwup_error;
-}
+static int n_arches_32 = sizeof(arch_names_32) / sizeof(arch_names_32[0]);
 
-#define EOKAY	0
-#define MAX_ERROR	35
+static char *arch_names_64[] = {
+#if defined(__x86_64__)
+	"x64",
+#elif defined(__aarch64__)
+	"aa64",
+#endif
+	""
+	};
 
-static const char const *error_table[MAX_ERROR - ERANGE] = {
-	[EOKAY] = "Okay",
-};
-
-static __thread char unknown[] = "Unknown error -2147483648";
-
-const char const *
-fwup_strerror(int error)
-{
-	if (error < 0 || error >= MAX_ERROR) {
-		sprintf(unknown, "Unknown error %d\n", error);
-		return unknown;
-	}
-	if (error > 0 && error <= ERANGE)
-		return strerror(error);
-	if (error == 0)
-		return dgettext("libfwup", error_table[error]);
-	return dgettext("libfwup", error_table[error - ERANGE]);
-}
-
-const char const *
-fwup_strerror_r(int error, char *buf, size_t buflen)
-{
-	if (!buf || !buflen) {
-		fwup_error = ERANGE;
-		return NULL;
-	}
-
-	size_t n = 0;
-	if (error < 0 || error >= MAX_ERROR) {
-		return strerror_r(error, buf, buflen);
-		n = snprintf(NULL, 0, "Unknown error %d", error);
-		if (n < buflen) {
-			fwup_error = EINVAL;
-			return NULL;
-		}
-		snprintf(buf, buflen, "Unknown error %d", error);
-		return buf;
-	}
-
-	if (error > 0 && error <= ERANGE)
-		return strerror_r(error, buf, buflen);
-
-	if (error == 0)
-		return dgettext("libfwup", error_table[error]);
-
-	n = strlen(dgettext("libfwup", error_table[error - ERANGE])) + 1;
-	if (n < buflen) {
-		fwup_error = EINVAL;
-		return NULL;
-	}
-	strcpy(buf, dgettext("libfwup", error_table[error - ERANGE]));
-	return buf;
-}
+static int n_arches_64 = sizeof(arch_names_64) / sizeof(arch_names_64[0]);
 
 #define ESRT_DIR "/sys/firmware/efi/esrt/"
 #define get_esrt_dir(entries)						\
@@ -167,15 +122,16 @@ get_info(efi_guid_t *guid, uint64_t hw_inst, update_info **info)
 	char *guidstr = NULL;
 	int rc;
 	update_info *local;
+	int error;
 
 	rc = efi_guid_to_str(guid, &guidstr);
 	if (rc < 0)
-		goto err;
+		return -1;
 	guidstr = onstack(guidstr, strlen(guidstr)+1);
 
 	rc = asprintf(&varname, "fwupdate-%s-%"PRIx64, guidstr, hw_inst);
 	if (rc < 0)
-		goto err;
+		return -1;
 	varname = onstack(varname, strlen(varname)+1);
 
 	uint8_t *data = NULL;
@@ -185,12 +141,10 @@ get_info(efi_guid_t *guid, uint64_t hw_inst, update_info **info)
 	rc = efi_get_variable(varguid, varname, &data, &data_size, &attributes);
 	if (rc < 0) {
 		if (errno != ENOENT)
-			goto err;
+			return -1;
 		local = calloc(1, sizeof (*local));
-		if (!local) {
-			rc = -1;
-			goto err;
-		}
+		if (!local)
+			return -1;
 
 		local->update_info_version = UPDATE_INFO_VERSION;
 		local->guid = *guid;
@@ -199,9 +153,10 @@ get_info(efi_guid_t *guid, uint64_t hw_inst, update_info **info)
 		local->dp_ptr = calloc(1, 1024);
 		if (!local->dp_ptr) {
 alloc_err:
-			fwup_error = errno;
+			error = errno;
 			free_info(local);
-			return rc;
+			errno = error;
+			return -1;
 		}
 
 		ssize_t sz;
@@ -214,29 +169,41 @@ alloc_err:
 		return 0;
 	}
 
-	if (data_size < sizeof (*local)) {
+	/* If our size is wrong, or our data is otherwise bad, try to delete
+	 * the variable and create a new one. */
+	if (data_size < sizeof (*local) || !data) {
+		if (data)
+			free(data);
 get_err:
-		errno = EINVAL;
-		free(local);
-		return -1;
+		rc = efi_del_variable(varguid, varname);
+		if (rc < 0)
+			return -1;
+		return get_info(guid, hw_inst, info);
 	}
 	local = (update_info *)data;
 
 	if (local->update_info_version != UPDATE_INFO_VERSION)
 		goto get_err;
 
-	efidp_header *dp = malloc(efidp_size((efidp)local->dp));
-	if (!dp)
-		goto get_err;
+	ssize_t sz = efidp_size((efidp)local->dp);
+	if (sz < 0) {
+		free(data);
+		errno = EINVAL;
+		return -1;
+	}
 
-	memcpy(dp, local->dp, efidp_size((efidp)local->dp));
+	efidp_header *dp = malloc((size_t)sz);
+	if (!dp) {
+		free(data);
+		errno = ENOMEM;
+		return -1;
+	}
+
+	memcpy(dp, local->dp, (size_t)sz);
 	local->dp_ptr = dp;
 
 	*info = local;
 	return 0;
-err:
-	fwup_error = errno;
-	return -1;
 }
 
 static int
@@ -246,12 +213,12 @@ put_info(update_info *info)
 	ssize_t dps, is;
 	char *guidstr = NULL;
 	char *varname;
+	int error;
 	int rc;
 
 	rc = efi_guid_to_str(&info->guid, &guidstr);
 	if (rc < 0) {
 err:
-		fwup_error = errno;
 		return rc;
 	}
 	guidstr = onstack(guidstr, strlen(guidstr)+1);
@@ -262,14 +229,24 @@ err:
 	varname = onstack(varname, strlen(varname)+1);
 
 	dps = efidp_size((efidp)info->dp_ptr);
-	is = (sizeof (*info)) + dps - (sizeof (info->dp_ptr));
-
-	update_info *info2;
-	info2 = alloca(is);
-	if (!info2) {
-		fwup_error = errno;
+	/* make sure dps is at least big enough to have our structure */
+	if (dps < 0 || (size_t)dps < sizeof(*info)) {
+		errno = EINVAL;
 		return -1;
 	}
+	/* Make sure sizeof(*info) + dps won't integer overflow */
+	if ((size_t)dps > SSIZE_MAX - sizeof(*info)) {
+		errno = EOVERFLOW;
+		return -1;
+	}
+
+	is = sizeof(*info) + dps - sizeof(info->dp_ptr);
+
+	update_info *info2;
+	info2 = malloc(is);
+	if (!info2)
+		return -1;
+
 	memcpy(info2, info, sizeof(*info));
 	memcpy(info2->dp, info->dp_ptr, dps);
 
@@ -277,8 +254,10 @@ err:
 			      | EFI_VARIABLE_BOOTSERVICE_ACCESS
 			      | EFI_VARIABLE_RUNTIME_ACCESS;
 	rc = efi_set_variable(varguid, varname, (uint8_t *)info2,
-			      is, attributes);
-	fwup_error = errno;
+			      is, attributes, 0644);
+	error = errno;
+	free(info2);
+	errno = error;
 	return rc;
 }
 
@@ -297,26 +276,29 @@ typedef struct fwup_resource_iter_s {
 int
 fwup_resource_iter_create(fwup_resource_iter **iter)
 {
+	int error;
 	if (!iter) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 	fwup_resource_iter *new = calloc(1, sizeof (fwup_resource_iter));
 	if (!new) {
-		fwup_error = ENOMEM;
+		errno = ENOMEM;
 		return -1;
 	}
 
 	new->dir = opendir(get_esrt_dir(1));
 	if (!new->dir) {
-		fwup_error = errno;
+err:
+		error = errno;
+		free(new);
+		errno = error;
 		return -1;
 	}
+
 	new->dirfd = dirfd(new->dir);
-	if (new->dirfd < 0) {
-		fwup_error = errno;
-		return -1;
-	}
+	if (new->dirfd < 0)
+		goto err;
 
 	*iter = new;
 	return 0;
@@ -337,7 +319,7 @@ int
 fwup_resource_iter_destroy(fwup_resource_iter **iterp)
 {
 	if (!iterp) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 	fwup_resource_iter *iter = *iterp;
@@ -358,7 +340,7 @@ fwup_resource_iter_next(fwup_resource_iter *iter, fwup_resource **re)
 {
 	fwup_resource *res;
 	if (!iter || !re) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 	res = &iter->re;
@@ -369,10 +351,8 @@ fwup_resource_iter_next(fwup_resource_iter *iter, fwup_resource **re)
 		errno = 0;
 		entry = readdir(iter->dir);
 		if (!entry) {
-			if (errno != 0) {
-				fwup_error = errno;
+			if (errno != 0)
 				return -1;
-			}
 			return 0;
 		}
 		if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, ".."))
@@ -381,7 +361,6 @@ fwup_resource_iter_next(fwup_resource_iter *iter, fwup_resource **re)
 
 	int dfd = openat(iter->dirfd, entry->d_name, O_RDONLY|O_DIRECTORY);
 	if (dfd < 0) {
-		fwup_error = errno;
 		return -1;
 	}
 
@@ -389,7 +368,6 @@ fwup_resource_iter_next(fwup_resource_iter *iter, fwup_resource **re)
 	get_string_from_file(dfd, "fw_class", &class);
 	int rc = efi_str_to_guid(class, &res->esre.guid);
 	if (rc < 0) {
-		fwup_error = errno;
 		return rc;
 	}
 	res->esre.fw_type = get_value_from_file(dfd, "fw_type");
@@ -403,10 +381,9 @@ fwup_resource_iter_next(fwup_resource_iter *iter, fwup_resource **re)
 			get_value_from_file(dfd, "lowest_supported_fw_version");
 
 	rc = get_info(&res->esre.guid, 0, &res->info);
-	if (rc < 0) {
-		fwup_error = errno;
+	if (rc < 0)
 		return rc;
-	}
+
 	res->info->capsule_flags = res->esre.capsule_flags;
 
 	*re = res;
@@ -418,7 +395,7 @@ int
 fwup_clear_status(fwup_resource *re)
 {
 	if (!re) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 
@@ -427,7 +404,6 @@ fwup_clear_status(fwup_resource *re)
 	re->info->status = 0;
 
 	rc = put_info(re->info);
-	fwup_error = errno;
 	return rc;
 }
 
@@ -435,7 +411,7 @@ int
 fwup_get_guid(fwup_resource *re, efi_guid_t **guid)
 {
 	if (!re || !guid) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 
@@ -447,7 +423,7 @@ int
 fwup_get_fw_version(fwup_resource *re, uint32_t *version)
 {
 	if (!re || !version) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 
@@ -459,7 +435,7 @@ int
 fwup_get_fw_type(fwup_resource *re, uint32_t *type)
 {
 	if (!re || !type) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 
@@ -471,7 +447,7 @@ int
 fwup_get_lowest_supported_fw_version(fwup_resource *re, uint32_t *version)
 {
 	if (!re || !version) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 
@@ -483,7 +459,7 @@ int
 fwup_get_attempt_status(fwup_resource *re, uint32_t *status)
 {
 	if (!re || !status) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 
@@ -497,12 +473,12 @@ fwup_get_last_attempt_info(fwup_resource *re, uint32_t *version,
 			   uint32_t *status, time_t *when)
 {
 	if (!re || !version || !status || !when) {
-		fwup_error = EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 
 	if (!re->info->status) {
-		fwup_error = ENOENT;
+		errno = ENOENT;
 		return -1;
 	}
 
@@ -530,46 +506,114 @@ fwup_get_last_attempt_info(fwup_resource *re, uint32_t *version,
 /* XXX PJFIX: this should be in efiboot-loadopt.h in efivar */
 #define LOAD_OPTION_ACTIVE      0x00000001
 
+static int
+get_paths(char **shim_fs_path, char **fwup_fs_path, char **fwup_esp_path)
+{
+	int ret = -1;
+
+	char shim_fs_path_tmpl[] = "/boot/efi/EFI/"FWUP_EFI_DIR_NAME"/shim";
+	char fwup_fs_path_tmpl[] = "/boot/efi/EFI/"FWUP_EFI_DIR_NAME"/fwup";
+	uint8_t fwup_esp_path_tmpl[] = "\\fwup";
+
+	char *shim_fs_path_tmp = NULL;
+	char *fwup_fs_path_tmp = NULL;
+	char *fwup_esp_path_tmp = NULL;
+
+	uint64_t firmware_bits = 0;
+
+	firmware_bits = get_value_from_file_at_dir("/sys/firmware/efi/",
+						   "fw_platform_size");
+	char **arch_names = firmware_bits == 64 ? arch_names_64
+						 : arch_names_32;
+	int n_arches = firmware_bits == 64 ? n_arches_64 : n_arches_32;
+	int i;
+
+	int rc;
+
+	*shim_fs_path = NULL;
+	*fwup_fs_path = NULL;
+	*fwup_esp_path = NULL;
+
+	i = find_matching_file(shim_fs_path_tmpl, ".efi", arch_names,
+			       n_arches, &shim_fs_path_tmp);
+
+	i = find_matching_file(fwup_fs_path_tmpl, ".efi", arch_names,
+				       n_arches, &fwup_fs_path_tmp);
+	if (i < 0) {
+		errno = ENOENT;
+		ret = i;
+		goto out;
+	}
+	rc = asprintf(&fwup_esp_path_tmp, "%s%s.efi", fwup_esp_path_tmpl,
+		      arch_names[i]);
+	if (rc < 0)
+		goto out;
+
+	if (shim_fs_path_tmp) {
+		*shim_fs_path = strdup(shim_fs_path_tmp);
+		if (!*shim_fs_path)
+			goto out;
+	}
+	if (fwup_fs_path_tmp) {
+		*fwup_fs_path = strdup(fwup_fs_path_tmp);
+		if (!*fwup_fs_path)
+			goto out;
+	}
+	if (fwup_esp_path_tmp)
+		*fwup_esp_path = fwup_esp_path_tmp;
+
+	return 0;
+out:
+	if (*shim_fs_path)
+		free(*shim_fs_path);
+	if (*fwup_fs_path)
+		free(*fwup_fs_path);
+	if (fwup_esp_path_tmp)
+		free(fwup_esp_path_tmp);
+	return ret;
+}
+
 static int
 set_up_boot_next(void)
 {
 	ssize_t sz, dp_size = 0;
 	uint8_t *dp_buf = NULL;
-	struct stat statbuf;
 	int rc;
 	int saved_errno;
-
-	char shim_fs_path[] = "/boot/efi/EFI/"FWUP_EFI_DIR_NAME"/shim.efi";
-	char fwup_fs_path[] = "/boot/efi/EFI/"FWUP_EFI_DIR_NAME"/fwupdate.efi";
-	uint8_t fwup_esp_path[] = " \\fwupdate.efi";
-	int use_fwup_path = 0;
+	int ret = -1;
 
 	uint16_t *loader_str = NULL;
 	size_t loader_sz = 0;
 
-	rc = stat(shim_fs_path, &statbuf);
-	if (rc < 0 && errno == ENOENT) {
+	char *shim_fs_path = NULL;
+	char *fwup_fs_path = NULL;
+	char *fwup_esp_path = NULL;
+	int use_fwup_path = 0;
+
+	char *label = NULL;
+
+	rc = get_paths(&shim_fs_path, &fwup_fs_path, &fwup_esp_path);
+	if (rc < 0)
+		return -1;
+
+	if (!shim_fs_path)
 		use_fwup_path = 1;
-	} else if (rc < 0) {
-		return rc;
-	}
 
 	sz = efi_generate_file_device_path(dp_buf, dp_size, use_fwup_path
 							    ? fwup_fs_path
 							    : shim_fs_path,
 					   EFIBOOT_OPTIONS_IGNORE_FS_ERROR|
-					   EFIBOOT_OPTIONS_IGNORE_PMBR_ERR|
 					   EFIBOOT_ABBREV_HD);
 	if (sz < 0)
-		return -1;
+		goto out;
 
 	dp_size=sz;
 	dp_buf = calloc(1, dp_size);
 	if (!dp_buf)
-		return -1;
+		goto out;
 
 	if (!use_fwup_path) {
-		loader_str = utf8_to_ucs2(fwup_esp_path, -1);
+		loader_str = utf8_to_ucs2((uint8_t *)fwup_esp_path, -1);
 		loader_sz = ucs2len(loader_str, -1) * 2;
 		if (loader_sz)
 			loader_sz += 2;
@@ -580,18 +624,21 @@ set_up_boot_next(void)
 							    ? fwup_fs_path
 							    : shim_fs_path,
 					   EFIBOOT_OPTIONS_IGNORE_FS_ERROR|
-					   EFIBOOT_OPTIONS_IGNORE_PMBR_ERR|
 					   EFIBOOT_ABBREV_HD);
 	if (sz != dp_size)
-		return -1;
+		goto out;
 
 	uint8_t *opt=NULL;
 	ssize_t opt_size=0;
 	uint32_t attributes = LOAD_OPTION_ACTIVE;
-	int ret = -1;
+
+	rc = asprintf(&label, "Linux-Firmware-Updater %s", fwup_esp_path);
+	if (rc < 0)
+		goto out;
+
 	sz = efi_loadopt_create(opt, opt_size, attributes,
 				  (efidp)dp_buf, dp_size,
-				  (uint8_t *)"Linux Firmware Updater",
+				  (uint8_t *)label,
 				  (uint8_t *)loader_str, loader_sz);
 	if (sz < 0)
 		goto out;
@@ -601,7 +648,7 @@ set_up_boot_next(void)
 	opt_size = sz;
 	sz = efi_loadopt_create(opt, opt_size, attributes,
 				  (efidp)dp_buf, dp_size,
-				  (uint8_t *)"Linux Firmware Updater",
+				  (uint8_t *)label,
 				  (uint8_t *)loader_str, loader_sz);
 	if (sz != opt_size)
 		goto out;
@@ -611,7 +658,7 @@ set_up_boot_next(void)
 	char *name = NULL;
 
 	uint32_t boot_next = 0x10000;
-	int found=0;
+	int found = 0;
 
 	uint8_t *var_data = NULL;
 	size_t var_data_size = 0;
@@ -674,16 +721,16 @@ do_next:
 		efi_loadopt_attr_set(loadopt, LOAD_OPTION_ACTIVE);
 		rc = efi_set_variable(*guid, name, var_data,
 				      var_data_size, attr);
-		ret = rc;
+		free(var_data);
+		if (rc < 0)
+			goto out;
 	} else {
 		char boot_next_name[] = "Boot####";
 		for (uint32_t value = 0; value < 0x10000; value++) {
 			int div = value / (sizeof(set_entries[0]) * 8);
 			int mod = value % (sizeof(set_entries[0]) * 8);
 
-			if (!set_entries[div]) {
-				boot_next = div * sizeof(set_entries[0]) * 8;
-			} else if (set_entries[div] & (1<<mod)) {
+			if (set_entries[div] & (1<<mod)) {
 				continue;
 			}
 			boot_next = value;
@@ -694,7 +741,8 @@ do_next:
 			goto out;
 
 		sprintf(boot_next_name, "Boot%04X", boot_next);
-		rc = efi_set_variable(*guid, boot_next_name, opt, opt_size,
+		rc = efi_set_variable(efi_guid_global, boot_next_name, opt,
+				      opt_size,
 				      EFI_VARIABLE_NON_VOLATILE |
 				      EFI_VARIABLE_BOOTSERVICE_ACCESS |
 				      EFI_VARIABLE_RUNTIME_ACCESS);
@@ -704,7 +752,7 @@ do_next:
 	}
 
 	uint16_t real_boot_next = boot_next;
-	rc = efi_set_variable(*guid, "BootNext",
+	rc = efi_set_variable(efi_guid_global, "BootNext",
 			      (uint8_t *)&real_boot_next, 2,
 			      EFI_VARIABLE_NON_VOLATILE |
 			      EFI_VARIABLE_BOOTSERVICE_ACCESS |
@@ -717,187 +765,433 @@ out:
 		free(dp_buf);
 	if (opt)
 		free(opt);
+	if (label)
+		free(label);
+	if (fwup_esp_path)
+		free(fwup_esp_path);
+	if (fwup_fs_path)
+		free(fwup_fs_path);
+	if (shim_fs_path)
+		free(shim_fs_path);
 
 	errno = saved_errno;
 	return ret;
 }
 
-int
-fwup_set_up_update(fwup_resource *re, uint64_t hw_inst, int infd)
+/**
+ * get_existing_media_path:
+ * @info: the #update_info
+ *
+ * Return a media path to use for the update which has already been used by
+ * this specific GUID.
+ *
+ * Returns: a media path, or %NULL if no such path exists.
+ */
+static char *
+get_existing_media_path(update_info *info)
 {
 	int rc;
+	char *relpath = NULL;
 	char *fullpath = NULL;
-	int fd = -1;
-	ssize_t sz;
-	off_t offset;
-	update_info *info = NULL;
-	uint8_t *dp_buf = NULL;
 	uint16_t *ucs2file = NULL;
 	uint16_t ucs2len = 0;
 
-	offset = lseek(infd, 0, SEEK_CUR);
+	/* never set */
+	if (!info->dp_ptr)
+		goto out;
+	if (efidp_end_entire(info->dp_ptr))
+		goto out;
 
-	rc = get_info(&re->esre.guid, 0, &info);
-	if (rc < 0) {
-		warn("get_info failed.\n");
-		goto err;
+	/* find UCS2 string */
+	const_efidp idp = (const_efidp)info->dp_ptr;
+	while (1) {
+		if (efidp_type(idp) == EFIDP_END_TYPE &&
+				efidp_subtype(idp) == EFIDP_END_ENTIRE)
+			break;
+		if (efidp_type(idp) != EFIDP_MEDIA_TYPE ||
+				efidp_subtype(idp) !=EFIDP_MEDIA_FILE) {
+			rc = efidp_next_node(idp, &idp);
+			if (rc < 0)
+				break;
+			continue;
+		}
+		ucs2file = (uint16_t *)((uint8_t *)idp + 4);
+		ucs2len = efidp_node_size(idp) - 4;
+		break;
 	}
 
-	if (info->dp_ptr && !efidp_end_entire(info->dp_ptr)) {
-		const_efidp idp = (const_efidp)info->dp_ptr;
-		while (1) {
-			if (efidp_type(idp) == EFIDP_END_TYPE &&
-					efidp_subtype(idp) == EFIDP_END_ENTIRE)
-				break;
-			if (efidp_type(idp) != EFIDP_MEDIA_TYPE ||
-					efidp_subtype(idp) !=EFIDP_MEDIA_FILE) {
-				rc = efidp_next_node(idp, &idp);
-				if (rc < 0)
-					break;
-				continue;
-			}
-			ucs2file = (uint16_t *)((uint8_t *)idp + 4);
-			ucs2len = efidp_node_size(idp) - 4;
-			break;
+	/* nothing found */
+	if (!ucs2file || ucs2len <= 0)
+		goto out;
+
+	/* convert to something sane */
+	relpath = ucs2_to_utf8(ucs2file, ucs2len / sizeof (uint16_t));
+	if (!relpath)
+		goto out;
+
+	/* convert '\' to '/' */
+	untilt_slashes(relpath);
+
+	/* build a complete path */
+	rc = asprintf(&fullpath, "/boot/efi%s", relpath);
+	if (rc < 0)
+		fullpath = NULL;
+
+out:
+	free(relpath);
+	return fullpath;
+}
+
+/**
+ * get_fd_and_media_path:
+ * @info: the #update_info
+ * @path: (out): the path
+ *
+ * Opens a suitable file descriptor and sets a media path to use for the update.
+ *
+ * Returns: a FD, or -1 for error
+ */
+static int
+get_fd_and_media_path(update_info *info, char **path)
+{
+	char *fullpath = NULL;
+	int fd = -1;
+	int rc;
+
+	/* look for an existing variable that we've used before for this
+	 * update GUID, and reuse the filename so we don't wind up
+	 * littering the filesystem with old updates */
+	fullpath = get_existing_media_path (info);
+	if (fullpath) {
+		fd = open(fullpath, O_CREAT|O_TRUNC|O_CLOEXEC|O_RDWR, 0600);
+		if (fd < 0) {
+			warn("open of %s failed", fullpath);
+			goto out;
 		}
-
-		if (!ucs2file || ucs2len <= 0)
-			goto new;
-
-		char *relpath = ucs2_to_utf8(ucs2file, ucs2len
-						       / sizeof (uint16_t));
-		if (!relpath)
-			goto new;
-		relpath = onstack(relpath, ucs2len / sizeof (uint16_t) + 1);
-
-		untilt_slashes(relpath);
-
-		rc = asprintf(&fullpath, "/boot/efi/%s", relpath);
-		if (rc < 0) {
-			free(relpath);
-			goto new;
-		}
-
-		rc = open(fullpath, O_CREAT|O_TRUNC|O_CLOEXEC|O_RDWR);
-		if (rc < 0) {
-			free(relpath);
-			goto new;
-		}
-		fd = rc;
-		rc = -1;
 	} else {
-new:
+		/* fall back to creating a new file from scratch */
 		rc = asprintf(&fullpath,
 			      "/boot/efi/EFI/%s/fw/fwupdate-XXXXXX.cap",
 			      FWUP_EFI_DIR_NAME);
 		if (rc < 0) {
 			warn("asprintf failed");
-			goto err;
+			goto out;
 		}
-
-		rc = mkostemps(fullpath, 4, O_CREAT|O_TRUNC|O_CLOEXEC|O_RDWR);
-		if (rc < 0) {
-			int saved_errno = errno;
+		fd = mkostemps(fullpath, 4, O_CREAT|O_TRUNC|O_CLOEXEC);
+		if (fd < 0) {
 			warn("mkostemps(%s) failed", fullpath);
-			free(fullpath);
-			errno = saved_errno;
-			goto err;
+			goto out;
 		}
-		fd = rc;
+	}
+
+	/* success, so take ownership of the string */
+	if (path) {
+		*path = fullpath;
+		fullpath = NULL;
+	}
+out:
+	free(fullpath);
+	return fd;
+}
+
+/**
+ * set_efidp_header:
+ * @info: the #update_info
+ * @path: the path
+ *
+ * Update the device path.
+ *
+ * Returns: a FD, or -1 for error
+ */
+static int
+set_efidp_header(update_info *info, const char *path)
+{
+	int rc = 0;
+	ssize_t req;
+	ssize_t sz;
+	uint8_t *dp_buf = NULL;
+
+	/* get the size of the path first */
+	req = efi_generate_file_device_path(NULL, 0, path,
+				EFIBOOT_OPTIONS_IGNORE_FS_ERROR |
+				EFIBOOT_ABBREV_HD);
+	if (req < 0) {
 		rc = -1;
+		goto out;
+	}
+	if (req <= 4) { /* if we just have an end device path,
+			  it's not going to work. */
+		rc = EINVAL;
+		goto out;
 	}
-	fullpath = onstack(fullpath, strlen(fullpath)+1);
 
+	dp_buf = calloc(1, req);
+	if (!dp_buf) {
+		rc = -1;
+		goto out;
+	}
+
+	/* actually get the path this time */
+	efidp_header *dp = (efidp_header *)dp_buf;
+	sz = efi_generate_file_device_path(dp_buf, req, path,
+				EFIBOOT_OPTIONS_IGNORE_FS_ERROR |
+				EFIBOOT_ABBREV_HD);
+	if (sz < 0) {
+		rc = -1;
+		goto out;
+	}
+
+	/* @info owns this now */
+	if (info->dp_ptr)
+		free(info->dp_ptr);
+	info->dp_ptr = dp;
+	dp_buf = NULL;
+out:
+	free(dp_buf);
+	return rc;
+}
+
+/**
+ * fwup_set_up_update
+ * @re: A %fwup_resource.
+ * @hw_inst: A hardware instance -- currently unused.
+ * @infd: file descriptor to the .cap binary
+ *
+ * Sets up a UEFI update using a file descriptor.
+ *
+ * Returns: -1 on error, @errno being set
+ *
+ * Since: 0.3
+ */
+int
+fwup_set_up_update(fwup_resource *re, uint64_t hw_inst, int infd)
+{
+	char *path = NULL;
+	int outfd = -1;
+	int rc;
+	off_t offset;
+	update_info *info = NULL;
+	FILE *fin = NULL, *fout = NULL;
+	int error;
+
+	/* check parameters */
+	if (infd < 0) {
+		warn("fd invalid.\n");
+		return -1;
+	}
+
+	offset = lseek(infd, 0, SEEK_CUR);
+
+	/* get device */
+	rc = get_info(&re->esre.guid, 0, &info);
+	if (rc < 0) {
+		warn("get_info failed.\n");
+		goto out;
+	}
+
+	/* get destination */
+	outfd = get_fd_and_media_path(info, &path);
+	if (outfd < 0) {
+		rc = -1;
+		goto out;
+	}
+
+	fin = fdopen(infd, "r");
+	if (!fin)
+		goto out;
+
+	fout = fdopen(outfd, "w");
+	if (!fout)
+		goto out;
+
+	/* copy the input file to the new home */
 	while (1) {
-		char buf[4096];
+		int c;
+		int rc;
 
-		sz = read(infd, &buf, 4096);
-		if (sz > 0) {
-			ssize_t wsz;
-			off_t off = 0;
-			while (sz-off) {
-				wsz = write(fd, buf+off, sz-off);
-				if (wsz < 0 &&
-				    (errno == EAGAIN || errno == EINTR))
-					continue;
-				if (wsz < 0) {
-					rc = wsz;
-					warn("write failed");
-					goto err;
-				}
-				off += wsz;
+		c = fgetc(fin);
+		if (c == EOF) {
+			if (feof(fin)) {
+				break;
+			} else if (ferror(fin)) {
+				warn("read failed");
+				rc = -1;
+				goto out;
+			} else {
+				warnx("fgetc() == EOF but no error is set.");
+				errno = EINVAL;
+				rc = -1;
+				goto out;
 			}
-			continue;
-		}
-		rc = sz;
-		if (sz < 0 && (errno == EAGAIN || errno == EINTR))
-			continue;
-		if (sz < 0) {
-			warn("read failed");
-			goto err;
-		}
-		if (sz == 0)
-			break;
-	}
-	close(infd);
-	infd = -1;
-	close(fd);
-	fd = -1;
-
-	if (!info->dp_ptr || efidp_end_entire(info->dp_ptr)) {
-		ssize_t req;
-		req = efi_generate_file_device_path(NULL, 0, fullpath,
-					EFIBOOT_OPTIONS_IGNORE_FS_ERROR|
-					EFIBOOT_OPTIONS_IGNORE_PMBR_ERR);
-		if (req < 0)
-			goto err;
-		if (req <= 4) { /* if we just have an end device path,
-				  it's not going to work. */
-			errno = EINVAL;
-			goto err;
 		}
 
-		dp_buf = calloc(1, req);
-		if (!dp_buf)
-			goto err;
-
-		efidp_header *dp = (efidp_header *)dp_buf;
-		sz = efi_generate_file_device_path(dp_buf, req, fullpath,
-					EFIBOOT_OPTIONS_IGNORE_FS_ERROR|
-					EFIBOOT_OPTIONS_IGNORE_PMBR_ERR);
-		if (sz < 0)
-			goto err;
-
-		if (info->dp_ptr)
-			free(info->dp_ptr);
-		info->dp_ptr = dp;
+		rc = fputc(c, fout);
+		if (rc == EOF) {
+			if (feof(fout)) {
+				break;
+			} else if (ferror(fout)) {
+				warn("write failed");
+				rc = -1;
+				goto out;
+			} else {
+				warnx("fputc() == EOF but no error is set.");
+				errno = EINVAL;
+				rc = -1;
+				goto out;
+			}
+		}
 	}
 
+	/* set efidp header */
+	rc = set_efidp_header(info, path);
+	if (rc < 0)
+		goto out;
+
+	/* save this to the hardware */
 	info->status = FWUPDATE_ATTEMPT_UPDATE;
-
+	memset(&info->time_attempted, 0, sizeof(info->time_attempted));
+	info->capsule_flags = re->esre.capsule_flags;
 	rc = put_info(info);
 	if (rc < 0) {
 		warn("put_info failed.\n");
-		goto err;
+		goto out;
 	}
 
-	free_info(info);
-
+	/* update the firmware before the bootloader runs */
 	rc = set_up_boot_next();
-	if (rc < 0) {
-		fwup_error = errno;
-		return rc;
-	}
-	return 1;
-err:
-	fwup_error = errno;
+	if (rc < 0)
+		goto out;
+out:
+	error = errno;
 	lseek(infd, offset, SEEK_SET);
-	if (dp_buf)
-		free(dp_buf);
-	if (info)
-		free_info(info);
-	if (fd > 0)
-		close(fd);
-
+	if (path)
+		free(path);
+	if (fin)
+		fclose(fin);
+	if (fout)
+		fclose(fout);
+	free_info(info);
+	if (outfd >= 0)
+		close(outfd);
+	errno = error;
 	return rc;
 }
+
+/**
+ * fwup_set_up_update_with_buf
+ * @re: A %fwup_resource.
+ * @hw_inst: A hardware instance -- currently unused.
+ * @buf: A memory buffer
+ * @sz: Size of @buf
+ *
+ * Sets up a UEFI update using a pre-allocated buffer.
+ *
+ * Returns: -1 on error, @errno being set
+ *
+ * Since: 0.5
+ */
+int
+fwup_set_up_update_with_buf(fwup_resource *re, uint64_t hw_inst, const void *buf, size_t sz)
+{
+	char *path = NULL;
+	int fd = -1;
+	int rc;
+	update_info *info = NULL;
+	int error;
+	off_t off = 0;
+
+	/* check parameters */
+	if (buf == NULL || sz == 0) {
+		warn("buf invalid.\n");
+		rc = -1;
+		goto out;
+	}
+
+	/* get device */
+	rc = get_info(&re->esre.guid, 0, &info);
+	if (rc < 0) {
+		warn("get_info failed.\n");
+		goto out;
+	}
+
+	/* get destination */
+	fd = get_fd_and_media_path(info, &path);
+	if (fd < 0) {
+		rc = -1;
+		goto out;
+	}
+
+	/* write the buf to a new file */
+	while (sz-off) {
+		ssize_t wsz;
+		wsz = write(fd, buf+off, sz-off);
+		if (wsz < 0 &&
+		    (errno == EAGAIN || errno == EINTR))
+			continue;
+		if (wsz < 0) {
+			rc = wsz;
+			warn("write failed");
+			goto out;
+		}
+		off += wsz;
+	}
+
+	/* set efidp header */
+	rc = set_efidp_header(info, path);
+	if (rc < 0)
+		goto out;
+
+	/* save this to the hardware */
+	info->status = FWUPDATE_ATTEMPT_UPDATE;
+	memset(&info->time_attempted, 0, sizeof(info->time_attempted));
+	info->capsule_flags = re->esre.capsule_flags;
+	rc = put_info(info);
+	if (rc < 0) {
+		warn("put_info failed.\n");
+		goto out;
+	}
+
+	/* update the firmware before the bootloader runs */
+	rc = set_up_boot_next();
+	if (rc < 0)
+		goto out;
+out:
+	error = errno;
+	free_info(info);
+	if (fd >= 0)
+		close(fd);
+	errno = error;
+	return rc;
+}
+
+/**
+ * fwup_last_attempt_status_to_string:
+ * @status: the status enum, e.g. %FWUP_LAST_ATTEMPT_STATUS_SUCCESS.
+ *
+ * Return a string representation of the last attempt status.
+ *
+ * Returns: A const string
+ *
+ * Since: 0.5
+ */
+const char *
+fwup_last_attempt_status_to_string (uint64_t status)
+{
+	if (status == FWUP_LAST_ATTEMPT_STATUS_SUCCESS)
+		return "Success";
+	if (status == FWUP_LAST_ATTEMPT_STATUS_ERROR_UNSUCCESSFUL)
+		return "Unsuccessful";
+	if (status == FWUP_LAST_ATTEMPT_STATUS_ERROR_INSUFFICIENT_RESOURCES)
+		return "Insufficient resources";
+	if (status == FWUP_LAST_ATTEMPT_STATUS_ERROR_INCORRECT_VERSION)
+		return "Incorrect version";
+	if (status == FWUP_LAST_ATTEMPT_STATUS_ERROR_INVALID_FORMAT)
+		return "Invalid firmware format";
+	if (status == FWUP_LAST_ATTEMPT_STATUS_ERROR_AUTH_ERROR)
+		return "Authentication signing error";
+	if (status == FWUP_LAST_ATTEMPT_STATUS_ERROR_PWR_EVT_AC)
+		return "AC power required";
+	if (status == FWUP_LAST_ATTEMPT_STATUS_ERROR_PWR_EVT_BATT)
+		return "Battery level is too low";
+	return NULL;
+}
